import pooler
import copy
from report import report_sxw

class evaluation_base_report(report_sxw.rml_parse):
    def __init__(self, cr, uid, name, context=None):
        super(evaluation_base_report, self).__init__(cr, uid, name, context)
        self.localcontext.update({
            'get_categories': self.get_categories,
            'get_level': self.get_level,
        })
        self.pool = pooler.get_pool(self.cr.dbname)

    def get_categories(self, obj_id):
        raise NotImplementedError

    def get_type(self):
        """need to be overloaded"""
        raise NotImplementedError

    def get_eval_domain(self, obj_id, results):
        """need to be overloaded"""
        raise NotImplementedError

    def hook_compute_categories_values(self, res, ev_lines, empty_categories, sorted_categories):
        """could be overloaded by child classes"""
        pass

    def get_level(self, category):
        level = 0
        while category and category.parent_id:
            level += 1
            category = category.parent_id
        return level

    def _get_avg(self, data, which=False):
        if which not in data:
            raise ValueError()

        if 'avg' not in data[which]:
            children = data[which].get('children', [])
            total = 0
            cnt = 0
            for child in children:
                val = self._get_avg(data, child)
                if val is not None:
                    total += val
                    cnt += 1

            v = data[which].get('value')
            if v is not None:
                v = sum(v) / len(v)
            if cnt:
                a = total / float(cnt)
                if v is not None:
                    a = (a + v) / 2
                data[which]['avg'] = a
            else:
                data[which]['avg'] = v

        return data[which]['avg']

    def _get_ordered_categories(self, proxy, results, sorted_categories, parent_id=False):
        search_domain = [('parent_id', '=', parent_id)]
        if self.get_type():
            search_domain.append(('type', '=', self.get_type()))

        ids = proxy.search(self.cr, self.uid, search_domain)
        results[parent_id]['children'] = ids
        for oid in ids:
            sorted_categories.append(oid)
            self._get_ordered_categories(proxy, results, sorted_categories, oid)

    def _compute_categories_values(self, ev_lines, empty_categories, sorted_categories):
        proxy = self.pool.get('training.evaluation.category')
        # duplicate empty results
        results = copy.deepcopy(empty_categories)

        # filter and update results dictionnary
        for ev_line in ev_lines:
            cat = ev_line.category_question_id.id
            if not results[cat]['value']:
                results[cat]['value'] = []
            results[cat]['value'].append(float(ev_line.response_value))

        # the calc average for those values
        self._get_avg(results)

        # sort results with sorted_categories order
        values = [(proxy.browse(self.cr, self.uid, cat_id), results[cat_id]['avg'],)
                    for cat_id in sorted_categories
                    if results[cat_id]['avg'] is not None]
        # if present, also append global result at the end
        if False in results:
            values.append((False, results[False]['avg'] or 0.0))
        return values

    def _count_related_evaluations(self, evaluation_line_ids):
        ev_count_query = (' SELECT COUNT(DISTINCT(evaluation_id))'
                         ' FROM training_evaluation_line'
                         ' WHERE id IN %s')
        if not evaluation_line_ids:
            return 0
        self.cr.execute(ev_count_query, (tuple(evaluation_line_ids),))
        ev_count_row = self.cr.fetchone() or (0,)
        return ev_count_row[0]


    def _get_categories(self, cr, uid, obj_id):
        results = { False : dict(value=None, children=list()) }
        sorted_categories = []
        categ_type = self.get_type()

        # for each category, setup default dict for it
        categ_domain = []
        if self.get_type():
            categ_domain.append(('type', '=', self.get_type()))
        proxy = self.pool.get('training.evaluation.category')
        for cat_id in proxy.search(cr, uid, []):
            results[cat_id] = dict(value=None, children=list())

        # fill result children and sorted_categories
        self._get_ordered_categories(proxy, results, sorted_categories)

        # get all corresponding evaluation.line
        ev_line_proxy = self.pool.get('training.evaluation.line')
        ids = ev_line_proxy.search(cr, uid, self.get_eval_domain(obj_id, results))

        # count number of evaluation related to thoses evaluation lines
        ev_count = self._count_related_evaluations(ids)
        res = {
            'number_evaluations': ev_count,
        }

        # get all evaluation lines, execute child classes hooks and get final global results
        ev_lines = ev_line_proxy.browse(cr, uid, ids)

        self.hook_compute_categories_values(res, ev_lines, results, sorted_categories)

        values = self._compute_categories_values(ev_lines, results, sorted_categories)
        global_avg = 0.0
        try:
            if values[-1][0] == False:
                global_avg = values[-1][1] or 0.0
        except IndexError:
            # dont fail if global value not present in list
            pass

        # finally return all that stuffs
        res.update({
            'categories' : values,
            'global_average' : global_avg,
        })
        return res

class evaluation_seance_report(evaluation_base_report):
    def get_categories(self, seance_id):
        return self._get_categories(self.cr, self.uid, seance_id)

    def get_type(self):
        return ''

    def get_eval_domain(self, obj_id, results):
        return [
            ('evaluation_id.evaluation_global_id.seance_id', '=', obj_id),
            ('category_question_id', 'in', results.keys()),
            ('type_question_id', '=', 'qcu'),
            ('response_value', '!=', False),
            ('response_value', '!=', 0),
        ]

report_sxw.report_sxw('report.training.evaluation.seance.report',
                      'training.seance',
                      'addons/training_evaluation/report/report_evaluation_seance.rml',
                      parser=evaluation_seance_report,
                      header=True)

class evaluation_global_report(evaluation_seance_report):
    def get_categories(self, global_id):
        glob = self.pool.get('training.evaluation.global').browse(self.cr, self.uid, global_id)
        return super(evaluation_global_report, self).get_categories(glob.seance_id.id)

report_sxw.report_sxw('report.training.evaluation.global.report',
                      'training.evaluation.global',
                      'addons/training_evaluation/report/report_evaluation_seance.rml',
                      parser=evaluation_global_report,
                      header=True)

class evaluation_organisation_report(evaluation_base_report):
    def get_categories(self, wizard_id):
        return self._get_categories(self.cr, self.uid, wizard_id)

    def get_type(self):
        return 'org'

    def get_eval_domain(self, obj_id, results):
        wizard = self.pool.get('training.evaluation.organization.report.wizard').browse(self.cr, self.uid, obj_id)
        return [
            ('category_question_partner_id', '=', wizard.partner_id.id),
            ('category_question_id', 'in', results.keys()),
            ('type_question_id', '=', 'qcu'),
            ('response_value', '!=', False),
            ('response_value', '!=', 0),
        ]

report_sxw.report_sxw('report.training.evaluation.organization.report',
                      'training.evaluation',
                      'addons/training_evaluation/report/report_evaluation_organization.rml',
                      parser=evaluation_organisation_report,
                      header=True)

class evaluation_contact_report(evaluation_base_report):
    def get_categories(self, contact_id):
        return self._get_categories(self.cr, self.uid, contact_id)

    def get_type(self):
        return 'lecturer'

    def get_eval_domain(self, obj_id, results):
        proxy_tpsh = self.pool.get('training.participation.stakeholder')
        ids = proxy_tpsh.search(self.cr, self.uid, [('contact_id', '=', obj_id),('state','in',['accepted','done'])])
        seance_ids = [ obj.seance_id.id for obj in proxy_tpsh.browse(self.cr, self.uid, ids) ]
        return [
            ('evaluation_id.evaluation_global_id.seance_id', 'in', seance_ids),
            ('category_question_id', 'in', results.keys()),
            ('type_question_id', '=', 'qcu'),
            ('response_value', '!=', False),
            ('response_value', '!=', 0),
        ]

    def hook_compute_categories_values(self, res, ev_lines, empty_categories, sorted_categories):
        ev_course_dict = {} # { 'course_id': evaluation line related}
        ev_course_info_dict = {} # {'course_id': { 'name': ..., ...}}
        course_proxy = self.pool.get('training.course')
        # get ev lines ids
        ev_lines_dict = dict([ (ev.id, ev) for ev in ev_lines ])
        ev_lines_ids = ev_lines_dict.keys()

        # get all distinct course for this evalutions lines
        ev_course_ids_q = (
            'SELECT tel.id AS id, tsea.course_id AS course_id, aaa.name AS course_name'
            ' FROM training_evaluation_line tel'
            ' LEFT JOIN training_evaluation te ON (tel.evaluation_id = te.id)'
            ' LEFT JOIN training_evaluation_global teg ON (te.evaluation_global_id = teg.id)'
            ' LEFT JOIN training_seance tsea ON (teg.seance_id = tsea.id)'
            ' LEFT JOIN training_course tc ON (tsea.course_id = tc.id)'
            ' LEFT JOIN account_analytic_account aaa ON (tc.analytic_account_id = aaa.id)'
            ' WHERE tel.id IN %s'
        )
        if ev_lines_dict:
            self.cr.execute(ev_course_ids_q, (tuple(ev_lines_ids),))
            for row in self.cr.dictfetchall():
                ev_course_info_dict.setdefault(row['course_id'], {})['name'] = row['course_name']
                ev_course_dict.setdefault(row['course_id'], []).append(ev_lines_dict[row['id']])

        # get standard course order
        ordered_course_ids = self.pool.get('training.course').search(self.cr, self.uid, [('id', 'in', ev_course_dict.keys())])

        # compute values + nb evals per courses and update final dict (res)
        if 'courses' not in res:
            res['courses'] = []
        for course_id in ordered_course_ids:
            course_values = self._compute_categories_values(ev_course_dict[course_id], empty_categories, sorted_categories)
            course_eval_cnt = self._count_related_evaluations([ ev_line.id for ev_line in ev_course_dict[course_id] ])
            res['courses'].append({
                'info': ev_course_info_dict[course_id],
                'course': course_proxy.browse(self.cr, self.uid, course_id),
                'values': course_values,
                'number_evaluations': course_eval_cnt
            })

report_sxw.report_sxw('report.training.evaluation.contact.report',
                      'res.partner.contact',
                      'addons/training_evaluation/report/report_evaluation_contact.rml',
                      parser=evaluation_contact_report,
                      header=True)

class contact_own_evaluation(report_sxw.report_sxw):
    def getObjects(self, cr, uid, ids, context=None):
        user_pool = pooler.get_pool(cr.dbname).get('res.users')
        user = user_pool.browse(cr, uid, uid, context=context)
        if user.contact_id:
            return [ user.contact_id ]
        return []

contact_own_evaluation('report.training.evaluation.contact.own.report',
                      'res.partner.contact',
                      'addons/training_evaluation/report/report_evaluation_contact.rml',
                      parser=evaluation_contact_report,
                      header=True)

class evaluation_course_report(evaluation_base_report):
    def get_categories(self, course_id):
        return self._get_categories(self.cr, self.uid, course_id)

    def get_type(self):
        return 'course'

    def get_eval_domain(self, obj_id, results):
        return [
            ('evaluation_id.evaluation_global_id.seance_id.course_id', '=', obj_id),
            ('category_question_id', 'in', results.keys()),
            ('type_question_id', '=', 'qcu'),
            ('response_value', '!=', False),
            ('response_value', '!=', 0),
        ]

    def hook_compute_categories_values(self, res, ev_lines, empty_categories, sorted_categories):
        ev_contact_dict = {} # { 'course_id': evaluation line related}
        contact_proxy = self.pool.get('res.partner.contact')
        # get ev lines ids
        ev_lines_dict = dict([ (ev.id, ev) for ev in ev_lines ])
        ev_lines_ids = ev_lines_dict.keys()

        # get all distinct course for this evalutions lines
        ev_lecturer_ids_q = (
            'SELECT tel.id AS id, tps.contact_id AS contact_id'
            ' FROM training_evaluation_line tel'
            ' LEFT JOIN training_evaluation te ON (tel.evaluation_id = te.id)'
            ' LEFT JOIN training_evaluation_global teg ON (te.evaluation_global_id = teg.id)'
            ' LEFT JOIN training_seance tsea ON (teg.seance_id = tsea.id)'
            ' LEFT JOIN training_participation_stakeholder tps ON (tps.seance_id = tsea.id AND tps.state IN (\'accepted\',\'done\'))'
            ' WHERE tel.id IN %s'
        )
        if ev_lines_ids:
            self.cr.execute(ev_lecturer_ids_q, (tuple(ev_lines_ids),))
            for row in self.cr.dictfetchall():
                ev_contact_dict.setdefault(row['contact_id'], []).append(ev_lines_dict[row['id']])

        # get standard course order
        ordered_contact_ids = self.pool.get('res.partner.contact').search(self.cr, self.uid, [('id', 'in', ev_contact_dict.keys())])

        # compute values + nb evals per courses and update final dict (res)
        if 'contacts' not in res:
            res['contacts'] = []
        for contact_id in ordered_contact_ids:
            contact_values = self._compute_categories_values(ev_contact_dict[contact_id], empty_categories, sorted_categories)
            contact_eval_cnt = self._count_related_evaluations([ ev_line.id for ev_line in ev_contact_dict[contact_id] ])
            res['contacts'].append({
                'contact': contact_proxy.browse(self.cr, self.uid, contact_id),
                'values': contact_values,
                'number_evaluations': contact_eval_cnt
            })

report_sxw.report_sxw('report.training.evaluation.course2.report',
                      'training.course',
                      'addons/training_evaluation/report/report_evaluation_course.rml',
                      parser=evaluation_course_report,
                      header=True)
